// Code generated by "stringer -type=Error"; DO NOT EDIT.

package types

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[ErrUndefined-0]
	_ = x[ErrEncode-1]
	_ = x[ErrDecode-2]
	_ = x[ErrClosed-3]
	_ = x[ErrTimeout-4]
	_ = x[ErrBadMessage-5]
	_ = x[ErrEmptyMessage-6]
	_ = x[ErrOversizedMessage-7]
	_ = x[ErrUnrecognizedMessage-8]
	_ = x[ErrPeerNotFound-9]
	_ = x[ErrBadAddress-10]
	_ = x[ErrBadKey-11]
}

const _Error_name = "ErrUndefinedErrEncodeErrDecodeErrClosedErrTimeoutErrBadMessageErrEmptyMessageErrOversizedMessageErrUnrecognizedMessageErrPeerNotFoundErrBadAddressErrBadKey"

var _Error_index = [...]uint8{0, 12, 21, 30, 39, 49, 62, 77, 96, 118, 133, 146, 155}

func (i Error) String() string {
	if i >= Error(len(_Error_index)-1) {
		return "Error(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _Error_name[_Error_index[i]:_Error_index[i+1]]
}
